@@ -192,9 +192,278 @@ static dsk_err_t check_density(LDBSDISK_DSK_DRIVER *self,
 }
 
 
+/*************************** DIAGNOSTIC DUMP FUNCTIONS ***********************/
+/* Because LDBS is handled by a library that I am trying to keep usable 
+ * separately from LibDsk, the diagnostic dump can't be done the normal way
+ * by instrumenting the file parser. Instead we do an entirely separate pass 
+ * over the file (after it's been loaded, so we know its structure is sound) -
+ * based on the reference utility ldbsdump.
+ *
+ * blocktype() expands the 4-character LDBS block type to something 
+ * hopefully human-readable.
+ */
+static char *blocktype(char *type)
+{
+	static char buf[40];
+
+	if (type[0] == 0 && type[1] == 0 && type[2] == 0 && type[3] == 0)
+		return "Free block";
+
+	if (!memcmp(type, LDBS_DIR_TYPE,     4)) return "Track directory";
+	if (!memcmp(type, LDBS_INFO_TYPE,    4)) return "Comment";
+	if (!memcmp(type, LDBS_DPB_TYPE,     4)) return "CP/M DPB";
+	if (!memcmp(type, LDBS_GEOM_TYPE,    4)) return "Drive geometry";
+	if (!memcmp(type, LDBS_CREATOR_TYPE, 4)) return "Creator";
+
+	if (type[0] == 'T')
+	{
+		sprintf(buf, "Track C%d H%d", ldbs_peek2((unsigned char *)type + 1), type[3]);
+		return buf;
+	}
+	if (type[0] == 'S')
+	{
+		sprintf(buf, "Sector C%d H%d S%d", type[1], type[2], type[3]);
+		return buf;
+	}
+
+	return NULL;
+}
+
+
+static unsigned char *load_block(DSK_REPORTFUNC diagfunc, FILE *fp, 
+		long pos, char *type, size_t *blocklen, size_t *datalen)
+{
+	unsigned char blockhead[20];
+	unsigned char *buf;
+	char *btype;
+
+	if (fseek(fp, pos, SEEK_SET))
+	{
+		return NULL;
+	}
+	if (fread(blockhead, 1, sizeof(blockhead), fp) < sizeof(blockhead) ||
+	    memcmp(blockhead, "LDB\1", 4))
+	{
+		return NULL;
+	}
+	*blocklen = ldbs_peek4(blockhead + 8);	
+	*datalen  = ldbs_peek4(blockhead + 12);	
+	buf = dsk_malloc(*blocklen);
+	if (!buf)
+	{
+		fseek(fp, pos + ldbs_peek4(blockhead + 8), SEEK_SET);
+		return NULL;
+	}
+	if (fread(buf, 1, *blocklen, fp) < *blocklen)
+	{
+		dsk_free(buf);
+		fseek(fp, pos + ldbs_peek4(blockhead + 8), SEEK_SET);
+		return NULL;
+	}
+	/* Seek to start of next block */
+	fseek(fp, pos + 20 + ldbs_peek4(blockhead + 8), SEEK_SET);
+	memcpy(type, blockhead + 4, 4);
+	btype = blocktype(type);
+
+	diaghead(diagfunc, "%s block", ldbs_peek4(blockhead + 4) ? "Data" : "Free");
+	diaghex(diagfunc, pos + 0, blockhead + 0, 4, "Block signature");
+	if (btype)
+	{
+		diaghex(diagfunc, pos + 4, blockhead + 4, 4, "Block type: %s", btype); 
+	}
+	else
+	{
+		diaghex(diagfunc, pos + 4, blockhead + 4, 4, "Block type");
+	}
+	diaghex(diagfunc, pos + 8, blockhead + 8, 4, "Block length");
+	diaghex(diagfunc, pos +12, blockhead +12, 4, "Block length used");
+	diaghex(diagfunc, pos +16, blockhead +16, 4, "Next block offset");
+
+	return buf;
+}
+
+static void dump_trackhead(DSK_REPORTFUNC diagfunc, long fpos,
+		unsigned char *trkhead, size_t datalen, int version)
+{
+	size_t n, count;
+	unsigned se_len, se_offset, th_offset;
+	long pos = fpos;
+
+	if (version < 2)
+	{
+		th_offset = 2;
+		se_offset = 6;
+		se_len    = 12;
+		count = ldbs_peek2(trkhead);
+
+		diaghex(diagfunc, pos, trkhead, 2, "Track header (%u sectors)",
+			count);
+		pos += 2;
+	}
+	else
+	{
+		th_offset = 6;
+		se_offset = ldbs_peek2(trkhead);
+		se_len    = ldbs_peek2(trkhead + 2);
+		count     = ldbs_peek2(trkhead + 4);
+		diaghex(diagfunc, pos,     trkhead, 2,     "Track header: Sectors at +%u", se_offset);
+		diaghex(diagfunc, pos + 2, trkhead + 2, 2, "Length of sector entries");
+		diaghex(diagfunc, pos + 4, trkhead + 4, 2, "%u sectors", count);
+		pos += 6;
+	}
+	diaghex(diagfunc, pos + 0, trkhead + th_offset + 0, 1, "Data rate");
+	diaghex(diagfunc, pos + 1, trkhead + th_offset + 1, 1, "Recording mode");
+	diaghex(diagfunc, pos + 2, trkhead + th_offset + 2, 1, "Format GAP3");
+	diaghex(diagfunc, pos + 3, trkhead + th_offset + 3, 1, "Filler byte");
+	pos += 4;
+	
+	if (se_offset >= 12)
+	{
+		diaghex(diagfunc, pos, trkhead + th_offset + 4, 2, "Total length");
+	}
+
+	pos = fpos + se_offset;
+	for (n = 0; n < count; n++)
+	{
+		diaghex(diagfunc, pos + se_len * n, 
+				  trkhead + se_offset + se_len * n,
+				  1, "Sector %d ID: Cylinder", n);
+		diaghex(diagfunc, pos + se_len * n + 1, 
+				  trkhead + se_offset + se_len * n + 1,
+				  1, "Sector %d ID: Head", n);
+		diaghex(diagfunc, pos + se_len * n + 2, 
+				  trkhead + se_offset + se_len * n + 2,
+				  1, "Sector %d ID: Sector number", n);
+		diaghex(diagfunc, pos + se_len * n + 3, 
+				  trkhead + se_offset + se_len * n + 3,
+				  1, "Sector %d ID: Sector size", n);
+		diaghex(diagfunc, pos + se_len * n + 4, 
+				  trkhead + se_offset + se_len * n + 4,
+				  1, "Sector %d: ST1", n);
+		diaghex(diagfunc, pos + se_len * n + 5, 
+				  trkhead + se_offset + se_len * n + 5,
+				  1, "Sector %d: ST2", n);
+		diaghex(diagfunc, pos + se_len * n + 6, 
+				  trkhead + se_offset + se_len * n + 6,
+				  1, "Sector %d: Copies", n);
+		diaghex(diagfunc, pos + se_len * n + 7, 
+				  trkhead + se_offset + se_len * n + 7,
+				  1, "Sector %d: Filler byte", n);
+		diaghex(diagfunc, pos + se_len * n + 8, 
+				  trkhead + se_offset + se_len * n + 8,
+				  4, "Sector %d: Data offset", n);
+
+		if (se_len >= 16)
+		{
+		    diaghex(diagfunc, pos + se_len * n + 12, 
+				  trkhead + se_offset + se_len * n + 12,
+				  2, "Sector %d: Trailing bytes", n);
+		    diaghex(diagfunc, pos + se_len * n + 14, 
+				  trkhead + se_offset + se_len * n + 14,
+				  2, "Sector %d: Offset", n);
+		}
+		if (se_len >= 18)
+		{
+		    diaghex(diagfunc, pos + se_len * n + 16, 
+				  trkhead + se_offset + se_len * n + 16,
+				  2, "Sector %d: Data length", n);
+		}
+	}
+}
+		
+
+static void dump_block(DSK_REPORTFUNC diagfunc, FILE *fp, long pos, int version)
+{
+	char type[4];
+	size_t blocklen, datalen;
+	unsigned char *data = load_block(diagfunc, fp, pos, type, &blocklen, &datalen);
+	int entries;
+	int n;
+	long fpos;
+	char *btype = blocktype(type);
+
+	if (data)
+	{
+		if (!memcmp(type, LDBS_DIR_TYPE, 4))
+		{
+			fpos = pos + 20;
+			entries = ldbs_peek2(data);
+			diaghex(diagfunc, fpos, data, 2, "Track directory: %d entries", entries);
+			for (n = 0; n < entries; n++)
+			{
+				char *p = blocktype((char *)data + 2 + 8 *n);
+
+				if (p) diaghex(diagfunc, fpos + 2 + 8 * n, data + 2 + 8 * n, 4, "Entry %d type: %s", n, p);
+				else  diaghex(diagfunc, fpos + 2 + 8 * n, data + 2 + 8 * n, 4, "Entry %d type", n);
+				diaghex(diagfunc, fpos + 6 + 8 * n, data + 6 + 8 * n, 4, "Entry %d offset", n);
+			}	
+			datalen = 2 + 8 * entries;
+			if (datalen < blocklen)	
+				diaghex(diagfunc, fpos + datalen, 
+					data + datalen, blocklen - datalen,
+					"Unused");
+		}
+		else if (type[0] == 'T') // Track header/
+		{
+			dump_trackhead(diagfunc, pos + 20, data, datalen,
+				version);	
+		}
+		else
+		{
+			if (btype) 
+			{
+				diaghex(diagfunc, pos + 20, data, datalen, "Data: %s", btype);
+			}
+			else   
+			{
+				diaghex(diagfunc, pos + 20, data, datalen, "Data");
+			}
+		}
+		if (datalen < blocklen)	
+		{
+			diaghex(diagfunc, pos + 20 + datalen, 
+				data + datalen, blocklen - datalen, "Unused");
+		}
+		dsk_free(data);	
+	}
+}
+
+
+
+/* Do a diagnostic dump (cf ldbsdump -raw) of an LDBS disk image.
+  The file has already been successfully parsed, so we know it has 
+  an LDBS structure */
+static void ldbsdisk_diagnose(DSK_REPORTFUNC diagfunc, FILE *fp)
+{
+	unsigned char header[20];
+	unsigned char blockhead[20];
+	long pos;
+	int version;
+
+	memset(header, 0, sizeof(header));
+	if (fread(header, 1, sizeof(header), fp) < sizeof(header)) return;
+
+	version = header[7];
+
+	diaghead(diagfunc, "LDBS disk image");
+	diaghex(diagfunc,  0, header,      4, "Magic number");	
+	diaghex(diagfunc,  4, header + 4,  4, "File type: version %d", version);	
+	diaghex(diagfunc,  8, header + 8,  4, "First used block");	
+	diaghex(diagfunc, 12, header + 12, 4, "First free block");	
+	diaghex(diagfunc, 16, header + 16, 4, "Track directory");	
+
+	pos = ftell(fp);
+	while (fread(blockhead, 1, sizeof(blockhead), fp) == sizeof(blockhead))
+	{
+		dump_block(diagfunc, fp, pos, version);
+		pos = ftell(fp);
+	}
+}
+
 
 /* Open DSK image, checking for the magic number */
-dsk_err_t ldbsdisk_open(DSK_DRIVER *pdriver, const char *filename)
+dsk_err_t ldbsdisk_open(DSK_DRIVER *pdriver, const char *filename,
+		DSK_REPORTFUNC diagfunc)
 {
 	LDBSDISK_DSK_DRIVER *self;
 	dsk_err_t err;
@@ -213,7 +482,15 @@ dsk_err_t ldbsdisk_open(DSK_DRIVER *pdriver, const char *filename)
 		ldbs_close(&self->ld_store);
 		return DSK_ERR_NOTME;
 	}
-
+	if (diagfunc)
+	{
+		FILE * fp = fopen(filename, "rb");
+		if (fp)
+		{
+			ldbsdisk_diagnose(diagfunc, fp);
+			fclose(fp);
+		}
+	}
 	return ldbsdisk_attach(pdriver);
 }
 
@@ -245,6 +522,10 @@ dsk_err_t ldbsdisk_attach(DSK_DRIVER *pdriver)
 	}
 	self->ld_cur_cyl = -1;
 	self->ld_cur_head = -1;
+	if (self->ld_cur_track != NULL)
+	{
+		dsk_free(self->ld_cur_track);
+	}
 	self->ld_cur_track = NULL;
 
 	return DSK_ERR_OK;
@@ -276,6 +557,10 @@ dsk_err_t ldbsdisk_creat(DSK_DRIVER *pdriver, const char *filename)
 	self->ld_sector = 0;
 	memset(&self->ld_dpb, 0, sizeof(self->ld_dpb));
 
+	if (self->ld_cur_track != NULL)
+	{
+		dsk_free(self->ld_cur_track);
+	}
 	self->ld_cur_track = NULL;
 	self->ld_cur_cyl = -1;
 	self->ld_cur_head = -1;
@@ -664,6 +949,10 @@ dsk_err_t ldbsdisk_xwrite(DSK_DRIVER *pdriver, const DSK_GEOMETRY *geom,
 
 		cursec->copies = 1;	/* We don't support multiple copies */
 		cursec->trail  = 0;	/* We don't support trailing bytes */
+/* [1.5.10] Update deleted-data flag */
+		if (deleted)
+			cursec->st2 |= 0x40;
+		else	cursec->st2 &= ~0x40;
 		err = ldbs_putblock(self->ld_store, &cursec->blockid, type,
 				buf, size_expect);
 	}
@@ -724,6 +1013,12 @@ dsk_err_t ldbsdisk_format(DSK_DRIVER *pdriver, DSK_GEOMETRY *geom,
 
 	/* The track's sectors have been removed. All that remains is 
 	 * its header (empty). Create a new header that will replace it. */
+	if (self->ld_cur_track)
+	{
+		dsk_free(self->ld_cur_track);
+		self->ld_cur_track = NULL;
+	}
+
 	self->ld_cur_track = ldbs_trackhead_alloc(geom->dg_sectors);
 	if (!self->ld_cur_track)
 	{
@@ -749,6 +1044,7 @@ dsk_err_t ldbsdisk_format(DSK_DRIVER *pdriver, DSK_GEOMETRY *geom,
 	{
 		case RECMODE_FM:  self->ld_cur_track->recmode = 1; break;
 		case RECMODE_MFM: self->ld_cur_track->recmode = 2; break;
+		case RECMODE_GCR_C64: self->ld_cur_track->recmode = RECMODE_GCR_C64; break;
 	}
 	self->ld_cur_track->gap3 = geom->dg_fmtgap;
 	self->ld_cur_track->filler = filler;
@@ -1054,7 +1350,7 @@ static char *option_names[] =
 dsk_err_t ldbsdisk_option_enum(DSK_DRIVER *self, int idx, char **optname)
 {
 	if (!self) return DSK_ERR_BADPTR;
-	if (self->dr_class != &dc_ldbsdisk) return DSK_ERR_BADPTR;
+	DC_CHECK(self);
 
 	if (idx >= 0 && idx < (int)MAXOPTION)
 	{
@@ -1071,8 +1367,9 @@ dsk_err_t ldbsdisk_option_set(DSK_DRIVER *self, const char *optname, int value)
 	LDBSDISK_DSK_DRIVER *ldbs_self;
 	unsigned idx;
 
+
 	if (!self || !optname) return DSK_ERR_BADPTR;
-	if (self->dr_class != &dc_ldbsdisk) return DSK_ERR_BADPTR;
+	DC_CHECK(self)
 	ldbs_self = (LDBSDISK_DSK_DRIVER *)self;
 
 	for (idx = 0; idx < MAXOPTION; idx++)
@@ -1114,7 +1411,7 @@ dsk_err_t ldbsdisk_option_get(DSK_DRIVER *self, const char *optname, int *value)
 	int v = 0;
 
 	if (!self || !optname) return DSK_ERR_BADPTR;
-	if (self->dr_class != &dc_ldbsdisk) return DSK_ERR_BADPTR;
+	DC_CHECK(self)
 	ldbs_self = (LDBSDISK_DSK_DRIVER *)self;
 
 	for (idx = 0; idx < MAXOPTION; idx++)
