@@ -64,10 +64,12 @@ DRV_CLASS dc_qrst =
 /* Expand a compressed QRST track: */
 
 static dsk_err_t expand(FILE *fp, unsigned char *trackbuf, size_t tracklen,
-			size_t compressed_len)
+			size_t compressed_len, DSK_REPORTFUNC diagfunc)
 {
 	int c, d;
 	size_t offset = 0;
+	unsigned char ch;
+	long fpos = ftell(fp);
 
 	while (compressed_len)
 	{
@@ -83,6 +85,8 @@ static dsk_err_t expand(FILE *fp, unsigned char *trackbuf, size_t tracklen,
 #endif
 			return DSK_ERR_NOTME;
 		}
+		ch = c;
+		diaghex(diagfunc, fpos, &ch, 1, "Literal run of %d bytes", ch);
 		--compressed_len; 
 		if (!compressed_len) 
 		{
@@ -113,10 +117,13 @@ static dsk_err_t expand(FILE *fp, unsigned char *trackbuf, size_t tracklen,
 #endif
 			return DSK_ERR_NOTME;
 		}
+		diaghex(diagfunc, fpos + 1, trackbuf + offset, c, 
+				"Literal data");
+		fpos += (c + 1);
 		compressed_len -= c;
 		offset += c;
 		if (!compressed_len) return DSK_ERR_OK;
-		c = fgetc(fp); 	/* Length of literal data */
+		c = fgetc(fp); 	/* Length of compressed data */
 		if (c == EOF) 
 		{
 #ifndef WIN16
@@ -124,10 +131,14 @@ static dsk_err_t expand(FILE *fp, unsigned char *trackbuf, size_t tracklen,
 #endif
 			return DSK_ERR_NOTME;
 		}
+		ch = c;
+		diaghex(diagfunc, fpos++, &ch, 1, "RLE run of %d bytes", ch);
 		--compressed_len; 
 		if (!compressed_len) return DSK_ERR_NOTME;
 		d = fgetc(fp); 	/* Byte to repeat */
 		--compressed_len; 
+		ch = d;
+		diaghex(diagfunc, fpos++, &ch, 1, "RLE pattern");
 		/* Unexpected EOF */
 		if (d == EOF) 
 		{
@@ -150,7 +161,7 @@ static dsk_err_t expand(FILE *fp, unsigned char *trackbuf, size_t tracklen,
 }
 
 
-dsk_err_t qrst_open(DSK_DRIVER *self, const char *filename)
+dsk_err_t qrst_open(DSK_DRIVER *self, const char *filename, DSK_REPORTFUNC diagfunc)
 {
 	DSK_GEOMETRY geom;
 	QRST_DSK_DRIVER *qrself;
@@ -199,6 +210,16 @@ dsk_err_t qrst_open(DSK_DRIVER *self, const char *filename)
 		fclose(fp);
 		return DSK_ERR_NOTME;	
 	}
+	diaghead(diagfunc, "Quick Release Sector Transfer disk image");
+	diaghex (diagfunc, 0, header,     5, "Magic number");
+	diaghex (diagfunc, 5, header + 5, 3, "Unknown");
+	diaghex (diagfunc, 8, header + 8, 4, "Checksum");
+	diaghex (diagfunc,12, header +12, 1, "Disk capacity");
+	diaghex (diagfunc,13, header +13, 1, "Volume %d", header[13]);
+	diaghex (diagfunc,14, header +14, 1, " of %d", header[14]);
+	diaghex (diagfunc,15, header +15,60, "Title");
+	diaghex (diagfunc,75, header +75, sizeof(header) - 75, "Description");
+
 	switch (header[12])
 	{
 		case 1: dg_stdformat(&geom, FMT_360K, NULL, NULL); break;
@@ -263,11 +284,17 @@ dsk_err_t qrst_open(DSK_DRIVER *self, const char *filename)
 	while (fread(chbuf, 1, sizeof(chbuf), fp) == sizeof(chbuf))
 	{
 		long pos;
+		unsigned char ch;
 
 		cylinder = chbuf[0];
 		head     = chbuf[1];
 		pos = ftell(fp);
 
+		diaghead(diagfunc, "Cylinder %d head %d", cylinder, head);
+		diaghex(diagfunc, pos - 3, chbuf, 1, "Cylinder");
+		diaghex(diagfunc, pos - 2, chbuf + 1, 1, "Head");
+		diaghex(diagfunc, pos - 1, chbuf + 2, 1, "Compression");
+
 		err = DSK_ERR_OK;
 		switch (chbuf[2])
 		{
@@ -279,6 +306,8 @@ dsk_err_t qrst_open(DSK_DRIVER *self, const char *filename)
 #endif
 					err = DSK_ERR_NOTME;
 				}
+				diaghex(diagfunc, pos, trackbuf, tracklen,
+						"Uncompressed data");
 				break;
 			case 1: /* Blank */
 				c = fgetc(fp);
@@ -289,6 +318,9 @@ dsk_err_t qrst_open(DSK_DRIVER *self, const char *filename)
 					fprintf(stderr, "QRST: Short type 1 block @ 0x%lx\n", pos);
 #endif
 				}
+				ch = c;
+				diaghex(diagfunc, pos, &ch, 1,
+						"Fill track with 0x%02x", ch);
 				memset(trackbuf, c, tracklen);
 				break;
 			case 2: /* Compressed */
@@ -299,7 +331,14 @@ dsk_err_t qrst_open(DSK_DRIVER *self, const char *filename)
 					fprintf(stderr, "QRST: Short type 2 block @ 0x%lx\n", pos);
 #endif
 				}
-				else	err = expand(fp, trackbuf, tracklen, chbuf[0] + 256 * chbuf[1]);
+				else	
+				{
+					diaghex(diagfunc, pos, chbuf, 2,
+						"Length of compressed data");
+					err = expand(fp, trackbuf, tracklen, 
+						chbuf[0] + 256 * chbuf[1],
+						diagfunc);
+				}
 				break;
 			default: /* Unsupported compression type */
 				err = DSK_ERR_NOTME;
